#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <math.h>

#include "npb-C.h"
#include "global.h"

/**
 * thread private variables
 *
 * dims, xstart, ystart, zstart, xend, yend, zend
 * np1, np2
 */

int layout_type;
int fftblock;
int fftblockpad;


void cfftz (int is, int m, int n,
            dcomplex x_arr[NX][FFTBLOCKPAD],
            dcomplex y_arr[NX][FFTBLOCKPAD], int MYTHREAD);
void fftz2 (int is, int l, int m, int n, int ny, int ny1,
            dcomplex u_arr[NX],
            dcomplex x_arr[NX][FFTBLOCKPAD],
            dcomplex y_arr[NX][FFTBLOCKPAD]);

void setup_globals(int _layout_type, int _fftblock, int _fftblockpad)
{
    layout_type = _layout_type;
    fftblock = _fftblock;
    fftblockpad = _fftblockpad;
}

/*--------------------------------------------------------------------
  c-------------------------------------------------------------------*/

void ipow46 (double a, int exp_1, int exp_2, double *result)
{
    /*--------------------------------------------------------------------
      -------------------------------------------------------------------*/

    /*--------------------------------------------------------------------
      c compute a^exponent mod 2^46
      c-------------------------------------------------------------------*/

    double dummy, q, r;
    int n, n2;
    int two_pow;

    /*--------------------------------------------------------------------
      c Use
      c   a^n = a^(n/2)*a^(n/2) if n even else
      c   a^n = a*a^(n-1)       if n odd
      c-------------------------------------------------------------------*/
    *result = 1;
    if( (exp_2 == 0) || (exp_1 == 0) )
        return;
    q = a;
    r = 1;
    n = exp_1;
    two_pow = 1;

    while (two_pow)
    {
        n2 = n / 2;
        if (n2 * 2 == n)
        {
            dummy = randlc (&q, q);
            n = n2;
        }
        else
        {
            n = n * exp_2;
            two_pow = 0;
        }
    }

    while (n > 1)
    {
        n2 = n / 2;
        if (n2 * 2 == n)
        {
            dummy = randlc (&q, q);
            n = n2;
        }
        else
        {
            dummy = randlc (&r, q);
            n--;
        }
    }

    dummy = randlc (&r, q);
    *result = r;
}

/*--------------------------------------------------------------------
  c-------------------------------------------------------------------*/

int ilog2 (int n)
{
    /*--------------------------------------------------------------------
      -------------------------------------------------------------------*/

    int nn, lg;

    if (n == 1)
    {
        return 0;
    }
    lg = 1;
    nn = 2;
    while (nn < n)
    {
        nn = nn << 1;
        lg++;
    }

    return lg;
}

/*--------------------------------------------------------------------
  c-------------------------------------------------------------------*/

void compute_indexmap (int d[3], int xstart[3], int ystart[3], int zstart[3],
                       double *twiddle_ptr, int MYTHREAD)
{
    /*--------------------------------------------------------------------
      -------------------------------------------------------------------*/

    /*--------------------------------------------------------------------
      c compute function from local (i,j,k) to ibar^2+jbar^2+kbar^2 
      c for time evolution exponent. 
      c-------------------------------------------------------------------*/

    int i, j, k, ii, ii2, jj, ij2, kk, kk2, kj2;
    double ap;
    int d0 = d[0];
    int d1 = d[1];
    int d2 = d[2];
    int xstart2 = xstart[2];
    int ystart2 = ystart[2];
    int zstart2 = zstart[2];

    /*--------------------------------------------------------------------
      c basically we want to convert the fortran indices 
      c   1 2 3 4 5 6 7 8 
      c to 
      c   0 1 2 3 -4 -3 -2 -1
      c The following magic formula does the trick:
      c mod(i-1+n/2, n) - n/2
      c-------------------------------------------------------------------*/
    ap = -4.0 * ALPHA * PI * PI;

    switch (layout_type) {
        case 0:
            for (k = 0; k < d2; k++) {
                kk = (k+1+zstart[2]-2+NZ/2)%NZ - NZ/2;
                kk2 = kk*kk;
                for (j = 0; j < d1; j++) {
                    jj = (j+1+ystart[2]-2+NY/2)%NY - NY/2;
                    kj2 = jj*jj+kk2;
                    for (i = 0; i < d0; i++) {
                        ii =  (i+1+xstart[2]-2+NX/2)%NX - NX/2;
                        twid(k,j,i) = exp( ap * ((double) (ii*ii + kj2)) ); 
                    }
                }
            }
            break;
        case 1:
            for (i = 0; i < d1; i++) {
                ii = (i + 1 + xstart[2] - 2 + NX / 2) % NX - NX / 2;
                ii2 = ii * ii;
                for (j = 0; j < d2; j++) {
                    jj = (j + 1 + ystart[2] - 2 + NY / 2) % NY - NY / 2;
                    ij2 = jj * jj + ii2;
                    for (k = 0; k < d0; k++)
                    {
                        kk = (k + 1 + zstart[2] - 2 + NZ / 2) % NZ - NZ / 2;
                        twid(j,i,k) = exp( ap * ((double) (kk*kk + ij2)) );
                    }
                }
            }
            break;
    }
}

/*--------------------------------------------------------------------
  -------------------------------------------------------------------*/


void fft_init (int n)
{
    /** -----------------------------------------------------------------------
     * compute the roots-of-unity array that will be used for subsequent FFTs. 
     * ----------------------------------------------------------------------- */

    int m, ku, i, j, ln;
    double t, ti;

    /** -----------------------------------------------------------------------
     * initialize the U array with sines and cosines in a manner that permits
     * stride one access at each FFT iteration.
     * ----------------------------------------------------------------------- */
    m = ilog2 (n);
    u[0].real = (double) m;
    u[0].imag = 0.0;
    ku = 1;
    ln = 1;

    for (j = 1; j <= m; j++)
    {
        t = PI / ln;

        for (i = 0; i <= ln - 1; i++)
        {
            ti = i * t;
            u[i + ku].real = cos (ti);
            u[i + ku].imag = sin (ti);
        }

        ku = ku + ln;
        ln = 2 * ln;
    }

}

/*--------------------------------------------------------------------
  -------------------------------------------------------------------*/

void cffts1_comp (int is, int d[3], 
                  dcomplex *x,
                  dcomplex *xout,
                  dcomplex y0[NX][FFTBLOCKPAD],
                  dcomplex y1[NX][FFTBLOCKPAD], int MYTHREAD) {
    int logd[3];
    int i, j, k, jj;
    int d0 = d[0];
    int d1 = d[1];
    int d2 = d[2];
    int idx;

    for (i = 0; i < 3; i++) {
        logd[i] = ilog2 (d[i]);
    }

    for (k = 0; k < d2; k++) {
        for (jj = 0; jj <= d1 - fftblock; jj += fftblock) {
#if (TIMERS_ENABLED==TRUE)
            timer_start(T_FFTCOPY);
#endif
            for (j = 0; j < fftblock; j++) {
                for (i = 0; i < d0; i++) {
                    idx = k*d1*d0 + (j+jj)*d0 + i;
                    y0[i][j].real = x[idx].real;
                    y0[i][j].imag = x[idx].imag;
                }
            }
#if (TIMERS_ENABLED==TRUE)
            timer_stop(T_FFTCOPY); 
#endif

#if (TIMERS_ENABLED==TRUE)
            timer_start(T_FFTLOW); 
#endif
            cfftz (is, logd[0], d0, y0, y1, MYTHREAD);
#if (TIMERS_ENABLED==TRUE)
            timer_stop(T_FFTLOW);
#endif

#if (TIMERS_ENABLED==TRUE)
            timer_start(T_FFTCOPY);
#endif
            for (j = 0; j < fftblock; j++) {
                for (i = 0; i < d0; i++) {
                    idx = k*d1*d0 + (j+jj)*d0 + i;
                    xout[idx].real = y0[i][j].real;
                    xout[idx].imag = y0[i][j].imag;
                }
            }
#if (TIMERS_ENABLED==TRUE)
            timer_stop(T_FFTCOPY);
#endif
        }
    }
}

void cffts2_comp (int is, int d[3], 
                  dcomplex *x,
                  dcomplex *xout,
                  dcomplex y0[NX][FFTBLOCKPAD],
                  dcomplex y1[NX][FFTBLOCKPAD], int MYTHREAD)
{
    /*--------------------------------------------------------------------
      c-------------------------------------------------------------------*/

    int d0 = d[0];
    int d1 = d[1];
    int d2 = d[2];
    int logd[3];
    int i, j, k, ii;
    int idx;

    for (i = 0; i < 3; i++)
    {
        logd[i] = ilog2 (d[i]);
    }

    for (k = 0; k < d2; k++)
    {
        for (ii = 0; ii <= d0 - fftblock; ii += fftblock)
        {
#if (TIMERS_ENABLED==TRUE)
            timer_start(T_FFTCOPY);
#endif
            for (j = 0; j < d1; j++)
            {
                for (i = 0; i < fftblock; i++)
                {
                    idx = k*d1*d0 + j*d0 + i+ii;
                    y0[j][i].real = x[idx].real;
                    y0[j][i].imag = x[idx].imag;
                }
            }
#if (TIMERS_ENABLED==TRUE)
            timer_stop(T_FFTCOPY); 
#endif

#if (TIMERS_ENABLED==TRUE)
            timer_start(T_FFTLOW); 
#endif
            cfftz (is, logd[1], d1, y0, y1, MYTHREAD);
#if (TIMERS_ENABLED==TRUE)
            timer_stop(T_FFTLOW); 
#endif

#if (TIMERS_ENABLED==TRUE)
            timer_start(T_FFTCOPY); 
#endif
            for (j = 0; j < d1; j++)
            {
                for (i = 0; i < fftblock; i++)
                {
                    idx = k*d1*d0 + j*d0 + i+ii;
                    xout[idx].real = y0[j][i].real;
                    xout[idx].imag = y0[j][i].imag;
                }
            }
#if (TIMERS_ENABLED==TRUE)
            timer_stop(T_FFTCOPY);
#endif
        }
    }
}

void cffts3_comp (int is, int d[3],
                  dcomplex *x,
                  dcomplex *xout,
                  dcomplex y0[NX][FFTBLOCKPAD],
                  dcomplex y1[NX][FFTBLOCKPAD], int MYTHREAD)
{
    /*--------------------------------------------------------------------
      c-------------------------------------------------------------------*/

    int d0 = d[0];
    int d1 = d[1];
    int d2 = d[2];
    int logd[3];
    int i, j, k, ii;
    int idx;

    for (i = 0; i < 3; i++)
    {
        logd[i] = ilog2 (d[i]);
    }

    for (j = 0; j < d1; j++)
    {
        for (ii = 0; ii <= d0 - fftblock; ii += fftblock)
        {
#if (TIMERS_ENABLED==TRUE)
            timer_start(T_FFTCOPY);
#endif
            for (k = 0; k < d2; k++)
            {
                for (i = 0; i < fftblock; i++)
                {
                    idx = k*d1*d0 + j*d0 + i+ii;
                    y0[k][i].real = x[idx].real;
                    y0[k][i].imag = x[idx].imag;
                }
            }
#if (TIMERS_ENABLED==TRUE)
            timer_stop(T_FFTCOPY);
#endif

#if (TIMERS_ENABLED==TRUE)
            timer_start(T_FFTLOW);
#endif
            cfftz (is, logd[2], d2, y0, y1, MYTHREAD);
#if (TIMERS_ENABLED==TRUE)
            timer_stop(T_FFTLOW);
#endif

#if (TIMERS_ENABLED==TRUE)
            timer_start(T_FFTCOPY);
#endif
            for (k = 0; k < d2; k++)
            {
                for (i = 0; i < fftblock; i++)
                {
                    idx = k*d1*d0 + j*d0 + i+ii;
                    xout[idx].real = y0[k][i].real;
                    xout[idx].imag = y0[k][i].imag;
                }
            }
#if (TIMERS_ENABLED==TRUE)
            timer_stop(T_FFTCOPY);
#endif
        }
    }
}

/*--------------------------------------------------------------------
  c-------------------------------------------------------------------*/

void cfftz (int is, int m, int n,
            dcomplex x_arr[NX][FFTBLOCKPAD],
            dcomplex y_arr[NX][FFTBLOCKPAD], int MYTHREAD)
{
    /*--------------------------------------------------------------------
      c-------------------------------------------------------------------*/

    /*--------------------------------------------------------------------
      c   Computes NY N-point complex-to-complex FFTs of X using an algorithm due
      c   to Swarztrauber.  X is both the input and the output array, while Y is a 
      c   scratch array.  It is assumed that N = 2^M.  Before calling CFFTZ to 
      c   perform FFTs, the array U must be initialized by calling CFFTZ with IS 
      c   set to 0 and M set to MX, where MX is the maximum value of M for any 
      c   subsequent call.
      c-------------------------------------------------------------------*/

    int i, j, l, mx;

    if (fftblock==0) exit(0);

    /*--------------------------------------------------------------------
      c   Check if input parameters are invalid.
      c-------------------------------------------------------------------*/
    mx = (int) (u[0].real);
    if ((is != 1 && is != -1) || m < 1 || m > mx)
    {
        printf ("CFFTZ: Either U has not been initialized, or else\n"
                "one of the input parameters is invalid%5d%5d%5d\n",
                is, m, mx);
        exit (1);
    }

    /*--------------------------------------------------------------------
      c   Perform one variant of the Stockham FFT.
      c-------------------------------------------------------------------*/
    for (l = 1; l <= m; l += 2)
    {
        fftz2 (is, l, m, n, fftblock, fftblockpad, u, x_arr, y_arr);
        if (l == m)
            break;
        fftz2 (is, l + 1, m, n, fftblock, fftblockpad, u, y_arr, x_arr);
    }

    /*--------------------------------------------------------------------
      c   Copy Y to X.
      c-------------------------------------------------------------------*/
    if (m % 2 == 1)
    {
        for (j = 0; j < n; j++)
        {
            for (i = 0; i < fftblock; i++)
            {
                x_arr[j][i].real = y_arr[j][i].real;
                x_arr[j][i].imag = y_arr[j][i].imag;
            }
        }
    }
}

/*--------------------------------------------------------------------
  c-------------------------------------------------------------------*/

void fftz2 (int is, int l, int m, int n, int ny, int ny1,
            dcomplex u_arr[NX],
            dcomplex x_arr[NX][FFTBLOCKPAD],
            dcomplex y_arr[NX][FFTBLOCKPAD])
{
    /*--------------------------------------------------------------------
      c-------------------------------------------------------------------*/

    /*--------------------------------------------------------------------
      c   Performs the L-th iteration of the second variant of the Stockham FFT.
      c-------------------------------------------------------------------*/

    int k, n1, li, lj, lk, ku, i, j, i11, i12, i21, i22;
    dcomplex u1;

    /*--------------------------------------------------------------------
      c   Set initial parameters.
      c-------------------------------------------------------------------*/

    n1 = n / 2;
    if (l - 1 == 0)
    {
        lk = 1;
    }
    else
    {
        lk = 2 << ((l - 1) - 1);
    }
    if (m - l == 0)
    {
        li = 1;
    }
    else
    {
        li = 2 << ((m - l) - 1);
    }
    lj = 2 * lk;
    ku = li;

    for (i = 0; i < li; i++)
    {
        i11 = i * lk;
        i12 = i11 + n1;
        i21 = i * lj;
        i22 = i21 + lk;
        if (is >= 1)
        {
            u1.real = u_arr[ku + i].real;
            u1.imag = u_arr[ku + i].imag;
        }
        else
        {
            u1.real = u_arr[ku + i].real;
            u1.imag = -u_arr[ku + i].imag;
        }

        /*--------------------------------------------------------------------
          c   This loop is vectorizable.
          c-------------------------------------------------------------------*/
        for (k = 0; k < lk; k++)
        {
            for (j = 0; j < ny; j++)
            {
                double x11real, x11imag;
                double x21real, x21imag;
                x11real = x_arr[i11 + k][j].real;
                x11imag = x_arr[i11 + k][j].imag;
                x21real = x_arr[i12 + k][j].real;
                x21imag = x_arr[i12 + k][j].imag;
                y_arr[i21 + k][j].real = x11real + x21real;
                y_arr[i21 + k][j].imag = x11imag + x21imag;
                y_arr[i22 + k][j].real = u1.real * (x11real - x21real)
                    - u1.imag * (x11imag - x21imag);
                y_arr[i22 + k][j].imag = u1.real * (x11imag - x21imag)
                    + u1.imag * (x11real - x21real);
            }
        }
    }
}

void evolve (double *twiddle_ptr,
             dcomplex *u0_ptr, 
             dcomplex *u1_ptr, 
             int d[3])
{
    /*--------------------------------------------------------------------
      c-------------------------------------------------------------------*/

    /*--------------------------------------------------------------------
      c evolve u0 -> u1 (t time steps) in fourier space
      c-------------------------------------------------------------------*/

    int i, j, k;

    for (k = 0; k < d[2]; k++)
    {
        for (j = 0; j < d[1]; j++)
        {
            for (i = 0; i < d[0]; i++)
            {
                u0(k,j,i).real *= twid(k,j,i);
                u0(k,j,i).imag *= twid(k,j,i);
                u1(k,j,i) = u0(k,j,i);
            }
        }
    }
}

void checksum_comp(dcomplex *chk, dcomplex *u1_ptr,int d[3],
                   int *xstart, int *ystart, int *zstart,
                   int *xend, int *yend, int *zend){

    int j, q, r, s;
    int ox, oy, oz;

    for (j = 1; j <= 1024; j++) {
        q = (j % NX) + 1;
        if (q >= xstart[0] && q <= xend[0]) {
            r = ((3 * j) % NY) + 1;
            if (r >= ystart[0] && r <= yend[0]) {
                s = ((5 * j) % NZ) + 1;
                if (s >= zstart[0] && s <= zend[0]) {
                    oz = s-zstart[0];
                    oy = r-ystart[0];
                    ox = q-xstart[0];
                    chk->real += u1_ptr[oz*d[1]*d[0]+oy*d[0]+ox].real; 
                    chk->imag += u1_ptr[oz*d[1]*d[0]+oy*d[0]+ox].imag; 
                }
            }
        }
    }

}



