#include <upc_relaxed.h>
#include "mupc_reduce.h"

#define RMIN(x,y) ((x <= y)? (x) : (y))
#define RMAX(x,y) ((x >= y)? (x) : (y))

#define MBUF_WORD 16       /* number of r8 words */
#define MBUF_SIZE 128      /* for 16*8 */

shared [MBUF_SIZE] char sh_reduce_buf[THREADS*MBUF_SIZE];


void mupc_reduce_r8(double *data_in, double *data_out, int nitems, 
                    int opr, int node)
{
   int i, ip;
   int myid, num_procs, rnode;

   shared double *sh_ub;
   double ub[MBUF_WORD];

   myid = MYTHREAD;
   num_procs = THREADS;
   rnode = (node < 0 || node == myid);

   /* Copy local data to the shared buffer */
   if (num_procs > 1) {
      sh_ub = (shared double *)&sh_reduce_buf[MBUF_SIZE * myid];
      upc_memput(sh_ub, data_in, sizeof(double)*nitems);
   }
   if ((data_in != data_out) && rnode) {
      for (i = 0; i < nitems; i++) {
         data_out[i] = data_in[i];
      }
   }

   /* Reduce from this node (rnode == 1) or skip (rnode == 0) */
   if (num_procs < 2)
      return;

   /* Ensure everyone got here */
   upc_barrier 15;

   /* Reduce data from the shared buffer */
   for (ip = 0; ip < num_procs; ip++) {

      /* skip my own processor, since it is already dealt */
      if (ip == myid || !rnode) continue;

      sh_ub = (shared double *)&sh_reduce_buf[MBUF_SIZE * ip];
      upc_memget(ub, sh_ub, sizeof(double)*nitems);
      switch (opr) {
      case SMP_OPR_SUM:
         for (i = 0; i < nitems; i++)
            data_out[i] = data_out[i] + ub[i];
	 break;
      case SMP_OPR_MIN:
         for (i = 0; i < nitems; i++)
            data_out[i] = RMIN(data_out[i], ub[i]);
	 break;
      case SMP_OPR_MAX:
         for (i = 0; i < nitems; i++)
            data_out[i] = RMAX(data_out[i], ub[i]);
	 break;
      }
   }

   /* Ensure no one pass this point before buffer is used */
   upc_barrier 16;
}

void mupc_reduce_i4(int *data_in, int *data_out, int nitems, 
                    int opr, int node)
{
   int i, ip;
   int myid, num_procs, rnode;

   shared int *sh_ub;
   int ub[MBUF_WORD*2];

   myid = MYTHREAD;
   num_procs = THREADS;
   rnode = (node < 0 || node == myid);

   /* Copy local data to the shared buffer */
   if (num_procs > 1) {
      sh_ub = (shared int *)&sh_reduce_buf[MBUF_SIZE * myid];
      upc_memput(sh_ub, data_in, sizeof(int)*nitems);
   }
   if ((data_in != data_out) && rnode) {
      for (i = 0; i < nitems; i++) {
         data_out[i] = data_in[i];
      }
   }

   /* Reduce from this node (rnode == 1) or skip (rnode == 0) */
   if (num_procs < 2)
      return;

   /* Ensure everyone got here */
   upc_barrier 17;

   /* Reduce data from the shared buffer */
   for (ip = 0; ip < num_procs; ip++) {

      /* skip my own processor, since it is already dealt */
      if (ip == myid || !rnode) continue;

      sh_ub = (shared int *)&sh_reduce_buf[MBUF_SIZE * ip];
      upc_memget(ub, sh_ub, sizeof(int)*nitems);
      switch (opr) {
      case SMP_OPR_SUM:
         for (i = 0; i < nitems; i++)
            data_out[i] = data_out[i] + ub[i];
	 break;
      case SMP_OPR_MIN:
         for (i = 0; i < nitems; i++)
            data_out[i] = RMIN(data_out[i], ub[i]);
	 break;
      case SMP_OPR_MAX:
         for (i = 0; i < nitems; i++)
            data_out[i] = RMAX(data_out[i], ub[i]);
	 break;
      }
   }

   /* Ensure no one pass this point before buffer is used */
   upc_barrier 18;
}
